program SPDIFWAV;
(*
 S/PDIF interface control program. Creates a WAV (PCM) file.
 A CS8412 decodes the serial S/PDIF signal into different digital streams.
 These digital streams are fed to shift registers to extract the 16-bit
 audio data. This audio data is transferred into a 2048 deep FIFO. This
 FIFO is read and controlled by this program.

 IRQ10 is activated when FIFI is half full.

 $300-$301  Read FIFO
 $302-$303  Read status/clear interrupt
            Status: bit 7 last channel read (H=left, L=right)
                    bit 2 full FIFO (L=active)
                    bit 1 half full FIFO (L=active)
                    bit 0 empty FIFO (L=active)
 $304-$307  Reset FIFO

 All access is through READS, do NOT write to these registers.
 All access is 16-bit wide.

 Note: because of hardware limitations the channel information bit
       is somewhat 'delayed'. If the status is read directly following
       a data read then the channel information bit can not be thrusted.
*)


(*
  Format of the wave file is as follows:
  'RIFF'                                (text)
  filesize-8                            (4 bytes)
  'WAVE'                                (text)
  'fmt '                                (text)
  infosize     (=$00000010)             (4 bytes)
  format       (PCM=$0001)              (2 bytes)
  channels     (stereo=$0002)           (2 bytes)
  samples/sec  (44100=$0000AC44)        (4 bytes)
  bytes/sec    (44100*4=$0002B110)      (4 bytes)
  bytes/sample (=$0004)                 (2 bytes)
  bits/sample  (=$0010)                 (2 bytes)
  'data'                                (text)
  datasize     (=filesize-44)           (4 bytes)
  ....sample data....


  Note: a 4 bytes number $12345678 is saved as  $78 $56 $34 $12
        a 2 bytes number $1234     is saved as  $34 $12

*)


uses Crt, Dos;

type
   HeaderType = array[0..43] of byte;

const
   Header : HeaderType =
            (ord('R'), ord('I'), ord('F'), ord('F'),
             $00, $00, $00, $00,
             ord('W'), ord('A'), ord('V'), ord('E'),
             ord('f'), ord('m'), ord('t'), ord(' '),
             $10, $00, $00, $00,
             $01, $00,
             $02, $00,
             $44, $AC, $00, $00,
             $10, $B1, $02, $00,
             $04, $00,
             $10, $00,
             ord('d'), ord('a'), ord('t'), ord('a'),
             $00, $00, $00, $00 );
   Buffers = 10;
   FIFOread   = $300;
   FIFOstatus = $302;
   FIFOreset  = $304;
   TheLength  = 15 * 1024;
   BufferL    = TheLength * 2;

type
   DataBuffer = array[0..TheLength-1] of word;         { Large buffer }

var
  OldVec      : pointer;
  TheFile     : file of DataBuffer;
  ByteFile    : file of byte;
  TheData1    : DataBuffer;
  TheData2    : DataBuffer;
  WhichBuf    : boolean;
  CheckBuf    : boolean;
  StoreIndex  : word;
  First       : boolean;
  ByteCount   : longint;
  FileName    : string;
  Loop        : integer;
  Size        : longint;
  AByte       : byte;

{$F+}
{ Interrupt entered at half FIFO is full. The FIFO is 2048 deep, so
  1024 reads are allowed. }
procedure NewIRQ10; interrupt;
begin
  repeat
    if WhichBuf then TheData2[StoreIndex] := portw[FIFOread]
                else TheData1[StoreIndex] := portw[FIFOread];
    inc(StoreIndex);                      { Next address }
    if StoreIndex=TheLength then          { Next buffer? }
      begin
        StoreIndex := 0;
        WhichBuf := not(WhichBuf);
      end;
  until (portw[FIFOstatus] and $01=0);    { Until buffer empty }
  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


{$F+}
{ Interrupt entered at first and second half FIFO full. These are used
  to synchronize on the left channel. Due to speed the channel information
  can not directly after reading the data. For this the FIFO is first
  emptied and then several data and status is read as they appear in the
  FIFO (the datarate thus determining the speed). }
procedure FirstIRQ10; interrupt;
var
  Loop   : word;
  Dummy  : word;
  Status : word;
  Loop2  : word;
begin
  Loop := 0;
  repeat
    Dummy := portw[FIFOread];
  until (portw[FIFOstatus] and $01=0);     { Until buffer empty }
  for Loop2 := 0 to 10 do
    begin
      repeat until (portw[FIFOstatus] and $01)<>0;
      Status := portw[FIFOstatus];
      Dummy := portw[FIFOread];
      repeat until (portw[FIFOstatus] and $01)<>0;
      inc(Loop);
    end;
  if ((Status and $80)<>0) and odd(Loop)         { Synchronize }
    then Dummy := portw[FIFOread];
  SetIntVec($72, Addr(NewIRQ10));   { Normal interrupt procedure }
  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


procedure SetInterrupts;
var
  Dummy: word;
begin
  GetIntVec($72, OldVec);
  SetIntVec($72, Addr(FirstIRQ10));
  Dummy := portw[FIFOreset];    { Reset FIFO }
  Dummy := portw[FIFOstatus];   { Reset IRQ }
  asm
    in   al, $A1                { Enable status slave }
    and  al, $FB                { IRQ 10 enable }
    out  $A1, al
  end;
end;


procedure RestoreInterrupts;
var
  Dummy : word;
begin
  asm
    in   al, $A1                { Enable status slave }
    or   al, $04                { IRQ 10 disable }
    out  $A1, al
  end;
  SetIntVec($72, OldVec);
end;


begin
  if ParamCount<1 then
    begin
      writeln('Please use destination filename as parameter.');
      exit;
    end;

  FileName   := ParamStr(1);
  First      := true;
  WhichBuf   := false;
  CheckBuf   := WhichBuf;
  StoreIndex := 0;
  ByteCount  := 0;

  writeln;
  writeln('Writing to '+FileName);
  write( ByteCount:10, ' bytes written');
  assign(TheFile, FileName);
  rewrite(TheFile);

  for Loop := 0 to 21 do
    begin
      TheData1[Loop]   := (Header[Loop*2]) or (Header[Loop*2+1] shl 8);
      inc(StoreIndex);
    end;

  SetInterrupts;

  repeat
    if WhichBuf<>CheckBuf then
      begin
        if CheckBuf then write(TheFile, TheData2)
                    else write(TheFile, TheData1);
        ByteCount := ByteCount + BufferL;
        write(#13, ByteCount:10);
        CheckBuf := not(CheckBuf);
      end;
  until KeyPressed;

  RestoreInterrupts;

  Close(TheFile);
  writeln;
  Assign(ByteFile, FileName);
  reset(ByteFile);

  Size := FileSize(ByteFile)-8;                 { Overwrite filesize }
  seek(ByteFile, 4);
  AByte := Size and $FF;
  write(ByteFile, AByte);
  AByte := (Size and $FF00) shr 8;
  write(ByteFile, AByte);
  AByte := (Size and $FF0000) shr 16;
  write(ByteFile, AByte);
  AByte := (Size and $FF000000) shr 24;
  write(ByteFile, AByte);

  Size := FileSize(ByteFile)-44;                { Overwrite dataesize }
  seek(ByteFile, 40);
  AByte := Size and $FF;
  write(ByteFile, AByte);
  AByte := (Size and $FF00) shr 8;
  write(ByteFile, AByte);
  AByte := (Size and $FF0000) shr 16;
  write(ByteFile, AByte);
  AByte := (Size and $FF000000) shr 24;
  write(ByteFile, AByte);
  close(ByteFile);
end.
