unit W_spdif;
(*
 S/PDIF interface control program. Creates a WAV (PCM) file.
 A CS8412 decodes the serial S/PDIF signal into different digital streams.
 These digital streams are fed to shift registers to extract the 16-bit
 audio data. This audio data is transferred into a 2048 deep FIFO. This
 FIFO is read and controlled by this program.

 IRQ10 is activated when FIFI is half full.

 $300-$301  Read FIFO
 $302-$303  Read status/clear interrupt
            Status: bit 7 last channel read (H=left, L=right)
                    bit 2 full FIFO (L=active)
                    bit 1 half full FIFO (L=active)
                    bit 0 empty FIFO (L=active)
 $304-$307  Reset FIFO

 All access is through READS, do NOT write to these registers.
 All access is 16-bit wide.

 Note: because of hardware limitations the channel information bit
       is somewhat 'delayed'. If the status is read directly following
       a data read then the channel information bit can not be thrusted.
*)

(*
  Format of the wave file is as follows:
  'RIFF'                                (text)
  filesize-8                            (4 bytes)
  'WAVE'                                (text)
  'fmt '                                (text)
  infosize     (=$00000010)             (4 bytes)
  format       (PCM=$0001)              (2 bytes)
  channels     (stereo=$0002)           (2 bytes)
  samples/sec  (44100=$0000AC44)        (4 bytes)
  bytes/sec    (44100*4=$0002B110)      (4 bytes)
  bytes/sample (=$0004)                 (2 bytes)
  bits/sample  (=$0010)                 (2 bytes)
  'data'                                (text)
  datasize     (=filesize-44)           (4 bytes)
  ....sample data....


  Note: a 4 bytes number $12345678 is saved as  $78 $56 $34 $12
        a 2 bytes number $1234     is saved as  $34 $12

*)


interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, MPlayer, Menus, ExtCtrls, Buttons;

type
  TSPDIFForm = class(TForm)
    RecordButton: TBitBtn;
    BytesPanel: TPanel;
    StatusText: TLabel;
    StatusLabel: TLabel;
    MediaPlayer1: TMediaPlayer;
    SaveDialog1: TSaveDialog;
    Image1: TImage;
    ExitButton: TButton;
    CheckPanel: TPanel;
    ResourceLabel: TLabel;
    CheckLabel: TLabel;
    FilenameLabel: TLabel;
    DestinationLabel: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RecordButtonClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ExitButtonClick(Sender: TObject);
  private
    { Private declarations }
    procedure StartRecording;
  public
    { Public declarations }
  end;

var
  SPDIFForm: TSPDIFForm;

implementation

{$R *.DFM}


type
   HeaderType = array[0..43] of byte;

const
   Header : HeaderType =
            (ord('R'), ord('I'), ord('F'), ord('F'),
             $00, $00, $00, $00,
             ord('W'), ord('A'), ord('V'), ord('E'),
             ord('f'), ord('m'), ord('t'), ord(' '),
             $10, $00, $00, $00,
             $01, $00,
             $02, $00,
             $44, $AC, $00, $00,
             $10, $B1, $02, $00,
             $04, $00,
             $10, $00,
             ord('d'), ord('a'), ord('t'), ord('a'),
             $00, $00, $00, $00 );
   Buffers = 10;
   FIFOread   = $300;
   FIFOstatus = $302;
   FIFOreset  = $304;
   TheLength  = 31 * 1024;
   BufferL    = TheLength * 2;

type
   DataBuffer = array[0..TheLength-1] of word;         { Large buffer }

var
  InterruptSet: boolean;
  OldVec      : pointer;
  TheFile     : file of DataBuffer;
  ByteFile    : file of byte;
  TheData1    : ^DataBuffer;
  TheData2    : ^DataBuffer;
  WhichBuf    : boolean;
  CheckBuf    : boolean;
  StoreIndex  : word;
  First       : boolean;
  ByteCount   : longint;
  FileName    : string;
  Loop        : integer;
  Size        : longint;
  AByte       : byte;
  Done        : boolean;
  Samples     : integer;           { Sample counter for testing purposes }

{$F+}
{ Interrupt entered at half FIFO is full. The FIFO is 2048 deep, so
  1024 reads are allowed. }
procedure NewIRQ10; interrupt;
begin
  Samples := 0;
  repeat
    if WhichBuf then TheData2^[StoreIndex] := portw[FIFOread]
                else TheData1^[StoreIndex] := portw[FIFOread];
    inc(Samples);                         { Sample counter }
    inc(StoreIndex);                      { Next address }
    if StoreIndex=TheLength then          { Next buffer? }
      begin
        StoreIndex := 0;
        WhichBuf := not(WhichBuf);
      end;
  until (portw[FIFOstatus] and $01=0);    { Until buffer empty }
  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


procedure SetIntVec(Vector: byte; PAddress: pointer); assembler;
asm
   push ds
   mov  ah, $25
   mov  al, Vector
   lds  dx, PAddress
   call Dos3Call
   pop  ds
end;


function GetIntVec(Vector: byte): pointer; assembler;
asm
   mov  ah, $35
   mov  al, Vector
   call Dos3Call
   mov  dx, es             { Function result in DX:AX }
   mov  ax, bx
end;


{$F+}
{ Interrupt entered at first and second half FIFO full. These are used
  to synchronize on the left channel. Due to speed the channel information
  can not directly after reading the data. For this the FIFO is first
  emptied and then several data and status is read as they appear in the
  FIFO (the datarate thus determining the speed). }
procedure FirstIRQ10; interrupt;
var
  Loop   : word;
  Dummy  : word;
  Status : word;
  Loop2  : word;
begin
  Loop := 0;
  repeat
    Dummy := portw[FIFOread];
  until (portw[FIFOstatus] and $01=0);     { Until buffer empty }
  for Loop2 := 0 to 10 do
    begin
      repeat until (portw[FIFOstatus] and $01)<>0;
      Status := portw[FIFOstatus];
      Dummy := portw[FIFOread];
      repeat until (portw[FIFOstatus] and $01)<>0;
      inc(Loop);
    end;
  if ((Status and $80)<>0) and odd(Loop)         { Synchronize }
    then Dummy := portw[FIFOread];
  SetIntVec($72, Addr(NewIRQ10));   { Normal interrupt procedure }
  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


procedure SetInterrupts;
var
  Dummy: word;
begin
  if InterruptSet then Exit;
  OldVec := GetIntVec($72);
  SetIntVec($72, Addr(FirstIRQ10));
  Dummy := portw[FIFOreset];    { Reset FIFO }
  Dummy := portw[FIFOstatus];   { Reset IRQ }
  asm
    in   al, $A1                { Enable status slave }
    and  al, $FB                { IRQ 10 enable }
    out  $A1, al
  end;
  InterruptSet := true;
end;


procedure RestoreInterrupts;
var
  Dummy : word;
begin
  asm
    in   al, $A1                { Enable status slave }
    or   al, $04                { IRQ 10 disable }
    out  $A1, al
  end;
  SetIntVec($72, OldVec);
  InterruptSet := false;
end;


procedure TSPDIFForm.StartRecording;
var
  AMsg: TMsg;
  AStream: TFileStream;
begin
  MediaPlayer1.Close;
  RecordButton.Caption := 'Stop     ';
  RecordButton.Refresh;
  BytesPanel.Visible := true;
  CheckPanel.Visible := true;
  BytesPanel.Refresh;
  CheckPanel.Refresh;

  New(TheData1);
  New(TheData2);
  FileName   := SaveDialog1.FileName;
  First      := true;
  WhichBuf   := false;
  CheckBuf   := WhichBuf;
  StoreIndex := 0;
  ByteCount  := 0;
  Done       := false;


  AStream := TFileStream.Create(FileName, fmCreate);

  SetInterrupts;
  for Loop := 0 to 21 do
    begin
      TheData1^[Loop]   := (Header[Loop*2]) or (Header[Loop*2+1] shl 8);
      inc(StoreIndex);
    end;

  repeat
    if WhichBuf<>CheckBuf then
      begin
        if CheckBuf then AStream.Write(TheData2^, SizeOf(TheData2^))
                    else AStream.Write(TheData1^, SizeOf(TheData1^));
        ByteCount := ByteCount + BufferL;
        ResourceLabel.Caption := format('%d%%', [round(((Samples-1024)*100)/1024)]);
        Statuslabel.Caption := format('%d', [ByteCount]);
        StatusLabel.Refresh;
        CheckBuf := not(CheckBuf);
        Application.ProcessMessages;
      end;
  until Done;
  RestoreInterrupts;

  if ByteCount<>AStream.Position
    then StatusLabel.Caption := 'STREAMSIZE ERROR';

  AStream.Free;

  Dispose(TheData2);
  Dispose(TheData1);
  AssignFile(ByteFile, FileName);
  reset(ByteFile);

  Size := FileSize(ByteFile)-8;                 { Overwrite filesize }
  seek(ByteFile, 4);
  AByte := Size and $FF;
  write(ByteFile, AByte);
  AByte := (Size and $FF00) shr 8;
  write(ByteFile, AByte);
  AByte := (Size and $FF0000) shr 16;
  write(ByteFile, AByte);
  AByte := (Size and $FF000000) shr 24;
  write(ByteFile, AByte);

  Size := FileSize(ByteFile)-44;                { Overwrite dataesize }
  seek(ByteFile, 40);
  AByte := Size and $FF;
  write(ByteFile, AByte);
  AByte := (Size and $FF00) shr 8;
  write(ByteFile, AByte);
  AByte := (Size and $FF0000) shr 16;
  write(ByteFile, AByte);
  AByte := (Size and $FF000000) shr 24;
  write(ByteFile, AByte);
  CloseFile(ByteFile);

  CheckPanel.Visible := false;
  BytesPanel.Visible := false;
end;


procedure TSPDIFForm.FormCreate(Sender: TObject);
begin
  InterruptSet := false;
  Done         := false;
end;


procedure TSPDIFForm.FormDestroy(Sender: TObject);
begin
  if InterruptSet then RestoreInterrupts;
end;


procedure TSPDIFForm.RecordButtonClick(Sender: TObject);
begin
  if InterruptSet then
    begin
      Done := true;
      RecordButton.Caption := 'Record';
    end
    else
    begin
      StartRecording;
      MediaPlayer1.Filename := SaveDialog1.FileName;
      MediaPlayer1.Open;
    end;
end;


procedure TSPDIFForm.FormActivate(Sender: TObject);
var
  DisplayStr: string;
begin
  if not(SaveDialog1.Execute) then Application.Terminate;
  DisplayStr := SaveDialog1.FileName;
  if DestinationLabel.Canvas.TextWidth(DisplayStr)>DestinationLabel.Width then
    begin
      while DestinationLabel.Canvas.TextWidth('~~'+DisplayStr)>DestinationLabel.Width
        do Delete(DisplayStr, 1, 1);
      DisplayStr := '~~' + DisplayStr;  
    end;
  DestinationLabel.Caption := DisplayStr;
end;


procedure TSPDIFForm.ExitButtonClick(Sender: TObject);
begin
  if InterruptSet then Done := true
                  else Application.Terminate;
end;

end.
