program SPDIF1;
(*
 S/PDIF interface control program.
 A CS8412 decodes the serial S/PDIF signal into different digital streams.
 These digital streams are fed to shift registers to extract the 16-bit
 audio data. This audio data is transferred into a 2048 deep FIFO. This
 FIFO is read and controlled by this program.

 IRQ10 is activated when FIFI is half full.

 $300-$301  Read FIFO
 $302-$303  Read status/clear interrupt
            Status: bit 7 last channel read (H=left, L=right)
                    bit 2 full FIFO (L=active)
                    bit 1 half full FIFO (L=active)
                    bit 0 empty FIFO (L=active)
 $304-$307  Reset FIFO

 All access is through READS, do NOT write to these registers.
 All access is 16-bit wide.
*)

uses Crt, Dos;

const
   Buffers = 10;
   FIFOread   = $300;
   FIFOstatus = $302;
   FIFOreset  = $304;
   TheLength  = 15 * 1024;
   BufferL    = TheLength * 2;

type
   DataBuffer = array[0..TheLength-1] of word;         { Large buffer }

var
  OldVec      : pointer;
  TheFile     : file of DataBuffer;
  TheData1    : DataBuffer;
  TheData2    : DataBuffer;
  WhichBuf    : boolean;
  CheckBuf    : boolean;
  StoreIndex  : word;
  First       : boolean;
  ByteCount   : longint;

{$F+}
{ Interrupt entered at half FIFO is full. The FIFO is 2048 deep, so
  1024 reads are allowed. }
procedure NewIRQ10; interrupt;
var
  Loop: integer;
  Status: word;
begin
  repeat
    if WhichBuf then TheData2[StoreIndex] := portw[FIFOread]
                else TheData1[StoreIndex] := portw[FIFOread];
    inc(StoreIndex);
    Status := portw[FIFOstatus];
    if ((StoreIndex and $01)<>0) and ((Status and $80)<>0)
      then dec(StoreIndex);               { Keep channel in sync }
                                          { Odd address  = left  }
                                          { Even address = right }
                                          { NOTE PREC. INCREMENT }
    if StoreIndex>=TheLength then
      begin
        StoreIndex := 0;
        WhichBuf := not(WhichBuf);
      end;
  until (Status and $01=0);
  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


procedure SetInterrupts;
var
  Dummy: word;
begin
  GetIntVec($72, OldVec);
  SetIntVec($72, Addr(NewIRQ10));
  asm
    in   al, $A1                { Enable status slave }
    and  al, $FB                { IRQ 10 enable }
    out  $A1, al
  end;
  Dummy := portw[FIFOreset];    { Reset FIFO }
  Dummy := portw[FIFOstatus];   { Reset IRQ }
end;


procedure RestoreInterrupts;
var
  Dummy : word;
begin
  asm
    in   al, $A1                { Enable status slave }
    or   al, $04                { IRQ 10 disable }
    out  $A1, al
  end;
  SetIntVec($72, OldVec);
end;


begin
  First      := true;
  WhichBuf   := false;
  CheckBuf   := WhichBuf;
  StoreIndex := 0;
  ByteCount  := 0;

  writeln;
  writeln('Writing to G:TEST.RAW');
  write( ByteCount:10, ' bytes written');
  assign(TheFile, 'G:TEST.RAW');
  rewrite(TheFile);

  SetInterrupts;

  repeat
    if WhichBuf<>CheckBuf then
      begin
        if CheckBuf then write(TheFile, TheData2)
                    else write(TheFile, TheData1);
        ByteCount := ByteCount + BufferL;
        write(#13, ByteCount:10);
        CheckBuf := not(CheckBuf);
      end;
  until KeyPressed;

  RestoreInterrupts;

  Close(TheFile);
  writeln;
end.
