program SPSBWAV2;
(*
 S/PDIF interface control program. Creates a WAV (PCM) file.
 Before actual recording starts the S/PDIF signal is transferred to
 a SoundBlaster (compatible) device for sound output.
 During recording the sound output is continued (SPSBWAV does not have
 a sound output during recording).
 Note that due to some slight deviations in the timing of the recording
 and the playback devices an occasional 'plop' can occur.

 A CS8412 decodes the serial S/PDIF signal into different digital streams.
 These digital streams are fed to shift registers to extract the 16-bit
 audio data. This audio data is transferred into a 2048 deep FIFO. This
 FIFO is read and controlled by this program.

 IRQ10 is activated when FIFI is half full.

 $300-$301  Read FIFO
 $302-$303  Read status/clear interrupt
            Status: bit 7 last channel read (H=left, L=right)
                    bit 2 full FIFO         (L=active)
                    bit 1 half full FIFO    (L=active)
                    bit 0 empty FIFO        (L=active)
 $304-$307  Reset FIFO

 All access is through READS, do NOT write to these registers.
 All access is 16-bit wide.

 Note: because of hardware limitations the channel information bit
       is somewhat 'delayed'. If the status is read directly following
       a data read then the channel information bit can not be thrusted.
*)


(*
  Format of the wave file is as follows:
  'RIFF'                                (text)
  filesize-8                            (4 bytes)
  'WAVE'                                (text)
  'fmt '                                (text)
  infosize     (=$00000010)             (4 bytes)
  format       (PCM=$0001)              (2 bytes)
  channels     (stereo=$0002)           (2 bytes)
  samples/sec  (44100=$0000AC44)        (4 bytes)
  bytes/sec    (44100*4=$0002B110)      (4 bytes)
  bytes/sample (=$0004)                 (2 bytes)
  bits/sample  (=$0010)                 (2 bytes)
  'data'                                (text)
  datasize     (=filesize-44)           (4 bytes)
  ....sample data....


  Note: a 4 bytes number $12345678 is saved as  $78 $56 $34 $12
        a 2 bytes number $1234     is saved as  $34 $12

*)
{$M 16000, 32768, 128000}
{$X+}   { Function results not necessarily used }

uses Crt, Dos,
     Memory,
     SbUtil,
     DSPUtil,
     DMAUtil,
     MIXUtil;

const
  DSP_Name = 'DSP';
  UseSize  = 16384;                 { Number of bytes to use }
  BufSize  = (UseSize div 16)*16;   { Blocks of 16 bytes required }
  BufSamples = BufSize div 2;       { Samples in buffer }
  Error    = $ffff;
  No_Error = 0;

type
   HeaderType = array[0..43] of byte;

const
   Header : HeaderType =
            (ord('R'), ord('I'), ord('F'), ord('F'),
             $00, $00, $00, $00,
             ord('W'), ord('A'), ord('V'), ord('E'),
             ord('f'), ord('m'), ord('t'), ord(' '),
             $10, $00, $00, $00,
             $01, $00,
             $02, $00,
             $44, $AC, $00, $00,
             $10, $B1, $02, $00,
             $04, $00,
             $10, $00,
             ord('d'), ord('a'), ord('t'), ord('a'),
             $00, $00, $00, $00 );
   NrBuffers  = 2;
   FIFOread   = $300;
   FIFOstatus = $302;
   FIFOreset  = $304;

type
   DataBuffer = array[0..BufSamples-1] of word;           { Large buffer }

var
  SBB         : SBBASE;
  uMemSize    : word;
  lpPlayBuffer: ^DataBuffer;                    { Play back buffer }
  lpRecBuffer : array[1..NrBuffers] of ^DataBuffer;  { Recording buffers }
  RecBuffer   : byte;                           { Active recording buffer }
  RecSave     : byte;                           { Buffer to save (0=none) }
  RecIndex    : word;                           { Recording buffer index }
  PlayIndex   : word;                           { Playback buffer index }

  OldVec      : pointer;
  TheFile     : file of DataBuffer;
  ByteFile    : file of byte;

  First       : boolean;
  ByteCount   : longint;
  FileName    : string;
  Loop        : integer;
  Size        : longint;
  AByte       : byte;
  PlaySound   : boolean;
  BufCnt      : byte;
  Dummy       : char;
  SbOk        : boolean;
  dspcnt      : byte;

{$F+}
{ Interrupt entered at half FIFO is full. The FIFO is 2048 deep, so
  a minimum of 1024 reads are allowed. }
procedure NewIRQ10; interrupt;
var
  SoundData: word;
  Loop: word;
begin
  Loop := 0;                              { Loop counter }
  repeat
    SoundData := portw[FIFOread];         { Read data from FIFO }
    lpPlayBuffer^[PlayIndex] := SoundData;    { For sound output }
    lpRecBuffer[RecBuffer]^[RecIndex] := SoundData;
    inc(PlayIndex);                      { Next address sound output }
    inc(RecIndex);
    if PlayIndex=(BufSamples div 2) then    { Half buffer sound output? }
      begin
        BufCnt := 1;
      end;
    if PlayIndex=BufSamples then            { End buffer sound output? }
      begin
        BufCnt := 0;
        PlayIndex := 0;
      end;
                                          { Now the recording part }
    if RecIndex=BufSamples then           { Next buffer recording? }
      begin
        RecIndex := 0;
        RecSave := RecBuffer;             { Indicate buffer to save }
        inc(RecBuffer);
        if RecBuffer>NrBuffers then RecBuffer := 1;
      end;
    inc(Loop);                            { Next }
    SoundData := portw[FIFOstatus];
  until (Loop=1024);                      { Until half buffer empty }

  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


{$F+}
{ Interrupt entered at first and second half FIFO full. These are used
  to synchronize on the left channel. Due to speed the channel information
  can not directly determined after reading the data. For this the FIFO is
  first emptied and then several data and status is read as they appear in
  the FIFO (the datarate thus determining the speed). }
procedure FirstIRQ10; interrupt;
var
  Loop   : word;
  Dummy  : word;
  Status : word;
  Loop2  : word;
begin
  Loop := 0;
  repeat
    Dummy := portw[FIFOread];
  until (portw[FIFOstatus] and $01=0);     { Until buffer empty }
  for Loop2 := 0 to 10 do
    begin
      repeat until (portw[FIFOstatus] and $01)<>0;
      Status := portw[FIFOstatus];
      Dummy := portw[FIFOread];
      repeat until (portw[FIFOstatus] and $01)<>0;
      inc(Loop);
    end;
  if ((Status and $80)<>0) and odd(Loop)         { Synchronize }
    then Dummy := portw[FIFOread];
  SetIntVec($72, Addr(NewIRQ10));  { Normal interrupt procedure }
  asm
    mov  al,  $20
    out  $A0, al                { EOI slave }
    out  $20, al                { EOI master }
  end;
end;
{$F-}


procedure SetInterrupts;
var
  Dummy: word;
begin
  GetIntVec($72, OldVec);
  SetIntVec($72, Addr(FirstIRQ10));
  Dummy := portw[FIFOreset];    { Reset FIFO }
  Dummy := portw[FIFOstatus];   { Reset IRQ }
  asm
    in   al, $A1                { Enable status slave }
    and  al, $FB                { IRQ 10 enable }
    out  $A1, al
  end;
end;


procedure RestoreInterrupts;
var
  Dummy : word;
begin
  asm
    in   al, $A1                { Enable status slave }
    or   al, $04                { IRQ 10 disable }
    out  $A1, al
  end;
  SetIntVec($72, OldVec);
end;


{$F+}
procedure DSPIrq(b: byte);
begin
  write(#13,PlayIndex:6);
end;
{$F-}


{ Setup DSP/DMA/IRQ and play with/without realtime recording. }
procedure StartPlay(lpBuffer : pointer; isMemSize : word; Rec: boolean);
var iEncore   : boolean;
    m,
    cBackup   : byte;
    Freq      : word;
    Stereo    : boolean;
  DSP_cmd,
  LowB,
  HighB       : byte;
  dsp4_mode   : byte;
  Activate    : boolean;

const
    Temp      : DSP4UserFunc = DspIrq;

begin
  dsp4_Mode := 0;
  Freq := 44100;
  Stereo := true;
  dsp_AdjustFrq( Freq, FALSE, Stereo );
  dsp_InitIrqHandler;              { Install DSP interrupt function }
  dsp_ClearBuffer( lpBuffer, isMemSize );
  BufCnt := 0;
  PlayIndex := 0;
  repeat until BufCnt = 1;
  repeat until BufCnt = 0;
  mix4_SetVolume( VOICE, 255, 255 );
                             { Program DMA channel for 16 bits }
  m := MODE_SINGLE or MODE_AUTOINIT or MODE_READ;
  dma_SetChannel( SBB.iDspDmaW, lpBuffer, isMemSize, m);
                                       { Set input/output frequency }
  dsp4_DACFrq( Freq );
            { Playback is specified by command  byte }
  DSP_cmd := DSP4_CMDDAC;

  DSP_cmd := DSP_cmd or DSP4_CMDAUTOINIT or DSP4_CMDFIFO;
  DSP_cmd := DSP_cmd or DSP4_CMD16DMA;
  dsp4_Mode := DSP4_MODESTEREO;
  dsp4_Mode := dsp4_Mode or DSP4_MODESIGNED;

        { Send command, mode and transfer length to DSP }
  dsp_Write( DSP_cmd );
  dsp_Write( dsp4_Mode );
  dsp_Write( LO( ( uMemSize div 4 ) - 1 ) );
  dsp_Write( HI( ( uMemSize div 4 ) - 1 ) );

  dsp_InitWaitForIRQ;                { Initialize wait for IRQ }


end;


{ Stop realtime sound. }
procedure StopPlay(lpBuffer : pointer; isMemSize : word; Rec: boolean);
begin
  dsp_ClearBuffer( lpBuffer, isMemSize );

      { HISpeed commands can only be interrupted by resetting }
  dsp_Write( DSP4_EXIT16DMA );     { Exit after current block }
  dsp_WaitForNextIRQ( DummyUserFunc, 0 );          { wait for end }
  dsp_RestoreIrqHandler;          { install old IRQ handler }
end;


begin

  for Loop := 1 to NrBuffers do
    begin
      lpRecBuffer[Loop] := MemAlloc(BufSize);
      if lpRecBuffer[Loop] = NIL then
        begin
          writeln('Not enough memory to be allocated.');
          exit;
        end;
    end;

  writeln('NOTE: due to timing differences in recording and realtime playback');
  writeln('      some interference can be heard in the realtime playback. This');
  writeln('      has no effect on the recording itself!');
  writeln('      During realtime playback only (no recording) some ''plops'' can');
  writeln('      be observed, while during recording this will be noise-''plops''.');
  if ParamCount<1 then
    begin
      writeln('No destination filename given, realtime sound only.');
    end
    else
    begin
      writeln('Destination filename given, first realtime sound.');
    end;

  SbOk       := true;
  First      := true;
  RecBuffer  := 1;
  RecIndex   := 0;
  RecSave    := 0;
  PlayIndex  := 0;
  ByteCount  := 0;
  PlaySound  := true;

  SetInterrupts;

  if sb_GetEnviron( SBB, GetEnv( 'BLASTER' ) ) <> No_Error then
    begin
      writeln( 'BLASTER environment variable not available' );
      SbOk := false;
    end;

  if dsp_SetBase( SBB, TRUE ) <> No_Error then
    begin
      Writeln( 'BLASTER environment variable error' );
      SbOk := false;
    end;

  if SbOk then
    begin
      if not dsp_CanStereo then
        begin
          writeln(' WARNING! Stereo mode NOT supported.' );
          SbOk := false;
        end;
      if (16 > dsp_MaxBits) and SbOk then
        begin
          writeln(' WARNING! 16 bit mode NOT supported!' );
          SbOk := false;
        end;

      if SbOk then
        begin
          uMemSize := BufSize;
          lpPlayBuffer := dma_AllocMem( uMemSize );    { Allocate DMA memory }
          if lpPlayBuffer <> NIL then
            begin
              while keypressed do Dummy := ReadKey;

              if ParamCount<1 then write('Press any key to stop')
                              else write('Press any key to start recording');
              StartPlay( lpPlayBuffer, uMemSize, FALSE );
            end
            else
            begin
              writeln('No DMA memory available');
              SbOk := false;
            end;
        end;
    end;
    while Keypressed do Dummy := ReadKey;
    dspcnt := 1;
    if SbOk then
      repeat
        if dsp_WaitForNextIRQ( DummyUserFunc, 0 ) then Writeln( 'Data lost!' );
        if dspcnt=0 then dspcnt := 1
                    else dspcnt := 0;
        if dspcnt<>bufcnt then       { If out of sync }
        begin
          dsp_write($D5);          { Pauze DMA }
          repeat
          until dspcnt=bufcnt;     { Wait for in sync }
          dsp_write($D6);          { Resume DMA }
        end;
      until KeyPressed;
    while Keypressed do Dummy := ReadKey;
    if not(SbOk) or (ParamCount<1) then
      begin
        if SbOk then
          begin
            StopPlay( lpPlayBuffer, uMemSize, FALSE );
            dma_Free( lpPlayBuffer );                       { Free up memory }
          end;
        RestoreInterrupts;
        for Loop := 1 to NrBuffers do FreeMem(lpRecBuffer[Loop], BufSize);
        exit;
      end;

  writeln;
  writeln('Recording started');
  writeln('Press any key to stop');

  RestoreInterrupts;

  FileName   := ParamStr(1);
  First      := true;
  RecBuffer  := 1;
  RecIndex   := 0;
  RecSave    := 0;
  PlayIndex  := 0;
  ByteCount  := 0;
  PlaySound  := false;

  writeln;
  writeln('Writing to '+FileName);
  write( ByteCount:10, ' bytes written');
  assign(TheFile, FileName);
  rewrite(TheFile);

  for Loop := 0 to 21 do
    begin
      lpRecBuffer[RecBuffer]^[Loop]   := (Header[Loop*2]) or (Header[Loop*2+1] shl 8);
      inc(RecIndex);
    end;

  SetInterrupts;

  repeat
    if RecSave<>0 then
      begin
        write(TheFile, lpRecBuffer[RecSave]^);
        ByteCount := ByteCount + BufSize;
        write(#13, ByteCount:10);
        RecSave := 0;
      end;
  until KeyPressed;

  StopPlay( lpPlayBuffer, uMemSize, FALSE );
  dma_Free( lpPlayBuffer );                       { Free up memory }
  RestoreInterrupts;

  Close(TheFile);
  writeln;
  Assign(ByteFile, FileName);
  reset(ByteFile);

  for Loop := 1 to NrBuffers do FreeMem(lpRecBuffer[Loop], BufSize);

  Size := FileSize(ByteFile)-8;                 { Overwrite filesize }
  seek(ByteFile, 4);
  AByte := Size and $FF;
  write(ByteFile, AByte);
  AByte := (Size and $FF00) shr 8;
  write(ByteFile, AByte);
  AByte := (Size and $FF0000) shr 16;
  write(ByteFile, AByte);
  AByte := (Size and $FF000000) shr 24;
  write(ByteFile, AByte);

  Size := FileSize(ByteFile)-44;                { Overwrite datasize }
  seek(ByteFile, 40);
  AByte := Size and $FF;
  write(ByteFile, AByte);
  AByte := (Size and $FF00) shr 8;
  write(ByteFile, AByte);
  AByte := (Size and $FF0000) shr 16;
  write(ByteFile, AByte);
  AByte := (Size and $FF000000) shr 24;
  write(ByteFile, AByte);
  close(ByteFile);

end.
